use std::collections::{HashMap, HashSet};

use evdev_rs::enums::{EventCode, EventType};

use crate::{
    device::Device,
    event_code_iters::{code_to_string, flatten},
    graph_types::{Descriptor, NodeDescriptor},
    id_types::NodeId,
};

pub struct VirtualDevice {
    name: String,
    node_id: Option<NodeId>,

    inputs: HashMap<EventType, Vec<EventCode>>,
    outputs: HashMap<EventType, Vec<EventCode>>,
}

impl VirtualDevice {
    pub fn new(name: impl ToString) -> Self {
        Self {
            name: name.to_string(),
            node_id: None,

            inputs: HashMap::new(),
            outputs: HashMap::new(),
        }
    }

    pub fn connect_id(&mut self, node_id: NodeId) {
        self.node_id = Some(node_id);
    }

    pub fn build_ports(&mut self, devices: &[&Device]) {
        let mut inputs: HashMap<EventType, HashSet<EventCode>> = HashMap::new();
        let mut outputs: HashMap<EventType, HashSet<EventCode>> = HashMap::new();

        for device in devices {
            for (event_type, new_codes) in device.events() {
                match event_type {
                    EventType::EV_KEY | EventType::EV_REL | EventType::EV_ABS => {
                        match inputs.get_mut(event_type) {
                            Some(codes) => {
                                for code in new_codes {
                                    codes.insert(*code);
                                }
                            }
                            None => {
                                let mut codes = HashSet::new();

                                for code in new_codes {
                                    codes.insert(*code);
                                }

                                inputs.insert(*event_type, codes);
                            }
                        }
                    }
                    EventType::EV_FF => match outputs.get_mut(event_type) {
                        Some(codes) => {
                            for code in new_codes {
                                codes.insert(*code);
                            }
                        }
                        None => {
                            let mut codes = HashSet::new();

                            for code in new_codes {
                                codes.insert(*code);
                            }

                            outputs.insert(*event_type, codes);
                        }
                    },

                    _ => (),
                }
            }
        }

        for (event_type, codes) in inputs {
            self.inputs.insert(event_type, codes.into_iter().collect());
        }

        for (event_type, codes) in outputs {
            self.outputs.insert(event_type, codes.into_iter().collect());
        }
    }

    pub fn node_id(&self) -> NodeId {
        self.node_id.unwrap()
    }

    pub fn to_descriptor(&self) -> NodeDescriptor {
        NodeDescriptor {
            op_name: self.name.clone(),
            inputs: flatten(
                self.inputs
                    .iter()
                    .map(|(_event_type, codes)| {
                        codes
                            .iter()
                            .map(|code| (code_to_string(code), Descriptor::Button))
                            .collect::<Vec<(String, Descriptor)>>()
                    })
                    .collect::<Vec<Vec<(String, Descriptor)>>>(),
            ),
            outputs: flatten(
                self.outputs
                    .iter()
                    .map(|(_event_type, codes)| {
                        codes
                            .iter()
                            .map(|code| (code_to_string(code), Descriptor::Button))
                            .collect::<Vec<(String, Descriptor)>>()
                    })
                    .collect::<Vec<Vec<(String, Descriptor)>>>(),
            ),
        }
    }
}
